package com.qq.taf.net;

import java.net.InetSocketAddress;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.Iterator;

import com.qq.taf.proxy.TafLoggerCenter;

/**
 * 接收器
 * @author fanzhang
 */
public class Acceptor implements Runnable {
	
	protected static final int SOCKET_BACKLOG=10240;
	protected static final int DEFAULT_IO_TIMEOUT_MILLIS=30000;
	protected Processor[] processors;
	protected Selector selector;
	protected boolean started;
	protected String name;
	private Filter filter;
	protected Handler handler;
	protected int netThreads;
	protected int handleThreads;
	protected TCPAcceptedCallBack acceptCallBack;
	
	/**
	 * @param name 名称
	 * @param filter 过滤器
	 * @param handler 处理器
	 * @param netThreads 网络IO线程数（>0：给定大小线程池、=0：默认大小线程池、<0：等同于0）
	 * @param handleThreads 业务处理线程数（>0：给定大小线程池、=0：默认自增复用线程池、<0：无线程池）
	 */
	public Acceptor(String name,Filter filter,Handler handler,int netThreads,int handleThreads) {
		try {
			selector=Selector.open();
			started=false;
			this.name=name;
			this.filter=filter;
			this.handler=handler;
			this.netThreads=netThreads;
			this.handleThreads=handleThreads;
			this.acceptCallBack = new TCPAcceptedCallBack();
		} catch(Throwable t) {
			Logger.log("acceptor init error",t);
			Exception.raise(t);
		}
	}
	
	/**
	 * 设置ip拒绝名单，如果名单不空，将自动开启IP拒绝开关
	 * @param denyList
	 */
	public Acceptor setAcceptedCallBack(TCPAcceptedCallBack callBack) {
		this.acceptCallBack = callBack;
		return this;
	}

	public void run() {
		try {
			int sessionId=0;
			while(true) {
				int n=selector.select();
				if(n>0) {
					Iterator<SelectionKey> readyKeys=selector.selectedKeys().iterator();
					while(readyKeys.hasNext()) {
						SelectionKey readyKey=readyKeys.next();
						ServerSocketChannel serverSocketChannel=(ServerSocketChannel)readyKey.channel();
						SocketChannel socketChannel=serverSocketChannel.accept();
						if(!acceptCallBack.canAccept(socketChannel)) {
							TafLoggerCenter.info("ConnectionIpForbidden " +socketChannel.socket().getInetAddress());
							socketChannel.close();
							readyKeys.remove();
							continue;
						}
						int processorNum=Math.abs(sessionId)%processors.length;
						int ioTimeoutMillis=(Integer)readyKey.attachment();
						processors[processorNum].scheduleRegister(socketChannel, sessionId,ioTimeoutMillis);
						sessionId++;
						readyKeys.remove();
					}
				}
			}
		} catch(Throwable t){
			Logger.log("acceptor run error",t);
		}
	}
	
	/**
	 * 绑定监听套接口（可以绑定多个套接口）
	 * @param bindAddress 本地套接口
	 * @param ioTimeoutMillis IO超时毫秒数（>0：给定超时时间、=0：默认超时时间、<0：永不超时）
	 * @return 
	 */
	public synchronized Acceptor bind(InetSocketAddress bindAddress,int ioTimeoutMillis) {
		try {
			if(started) Exception.raise("can not bind any address after starting");
			ServerSocketChannel serverSocketChannel=ServerSocketChannel.open();
			serverSocketChannel.configureBlocking(false);
			serverSocketChannel.socket().setReuseAddress(true);
			serverSocketChannel.socket().bind(bindAddress,SOCKET_BACKLOG);
			SelectionKey key=serverSocketChannel.register(selector, SelectionKey.OP_ACCEPT);
			key.attach(ioTimeoutMillis!=0?ioTimeoutMillis:DEFAULT_IO_TIMEOUT_MILLIS);
		} catch(Throwable t) {
			Logger.log("acceptor bind error",t);
			Exception.raise(t);
		}
		return this;
	}
	
	/**
	 * 启动接收器
	 */
	public synchronized void listen() {
		try {
			started=true;
			
			HandlerExecutor handlerExecutor=handleThreads>=0?new HandlerExecutor(name,handleThreads,handler):new HandlerPseudoExecutor(handler);
			SessionWriter filterEncodeProduct=new SessionWriter();
			MessageReceiver filterDecodeProduct=new MessageReceiver(handlerExecutor);
			
			int processorCount=netThreads>0?netThreads:Runtime.getRuntime().availableProcessors()+1;
			processors=new Processor[processorCount];
			for(int i=0;i<processorCount;i++) processors[i]=new Processor(name,filter,filterEncodeProduct,filterDecodeProduct,handlerExecutor,i);
	
			new Thread(this,name+"-Acceptor").start();
		} catch(Throwable t) {
			Logger.log("acceptor listen error",t);
			Exception.raise(t);
		}
	}
	
}
