package com.qq.taf.net;

import java.net.InetSocketAddress;
import java.nio.channels.SocketChannel;


/**
 * 连接器
 * @author fanzhang
 */
public class Connector {
	
	protected static final int DEFAULT_IO_TIMEOUT_MILLIS=30000;
	protected Processor[] processors;
	protected SessionIdGenerator sessionIdGenerator;
	protected boolean started;
	protected String name;
	private Filter filter;
	protected Handler handler;
	protected int netThreads;
	protected int handleThreads;
	
	/**
	 * @param name 名称
	 * @param filter 过滤器
	 * @param handler 处理器
	 * @param netThreads 网络IO线程数（>0：给定大小线程池、=0：默认大小线程池、<0：等同于0）
	 * @param handleThreads 业务处理线程数（>0：给定大小线程池、=0：默认自增复用线程池、<0：无线程池）
	 */
	public Connector(String name,Filter filter,Handler handler,int netThreads,int handleThreads) {
		started=false;
		this.name=name;
		this.filter=filter;
		this.handler=handler;
		this.netThreads=netThreads;
		this.handleThreads=handleThreads;
		sessionIdGenerator=new SessionIdGenerator();
	}
	
	/**
	 * <p>注入一个SessionId生成器
	 * <p>说明：仅当多个Connector共用一套Filter与Handler，并且需要使用connect方法返回的SessionId的情况下，必须同时注入一个共用的SessionIdGenerator实例以确保SessionId被生成时的唯一性。
	 * @param sessionIdGenerator
	 */
	public void setSessionIdGenerator(SessionIdGenerator sessionIdGenerator) {
		this.sessionIdGenerator=sessionIdGenerator;
	}
	
	/**
	 * 连接到远程套接口
	 * <p> SessionId的创建由SessionIdGenerator负责
	 * @param connectAddress 远程套接口
	 * @param ioTimeoutMillis IO超时毫秒数（>0：给定超时时间、=0：默认超时时间、<0：永不超时）
	 * @return 新创建的SessionId
	 */
	public int connect(InetSocketAddress connectAddress,int ioTimeoutMillis) {
		int sessionId=sessionIdGenerator.newSessionId();
		connect(connectAddress, ioTimeoutMillis, sessionId);
		return sessionId;
	}
	
	/**
	 * 连接到远程套接口
	 * <p> 自行创建SessionId时使用之
	 * @param connectAddress 远程套接口
	 * @param ioTimeoutMillis IO超时毫秒数（>0：给定超时时间、=0：默认超时时间、<0：永不超时）
	 * @param sessionId 新创建的SessionId
	 */
	public void connect(InetSocketAddress connectAddress,int ioTimeoutMillis,int sessionId) {
		init();
		try {
			SocketChannel socketChannel=SocketChannel.open();
			socketChannel.socket().connect(connectAddress, ioTimeoutMillis>0?ioTimeoutMillis:DEFAULT_IO_TIMEOUT_MILLIS);
			int processorNum=Math.abs(sessionId)%processors.length;
			processors[processorNum].scheduleRegister(socketChannel,sessionId,ioTimeoutMillis!=0?ioTimeoutMillis:DEFAULT_IO_TIMEOUT_MILLIS); 
		} catch(Throwable t) {
			Logger.log("connector connect error, addr="+connectAddress,t);
			Exception.raise(t);
		}
	}
	
	protected void init() {
		if(!started) {
			synchronized (this) {
				if(!started) {
					HandlerExecutor handlerExecutor=handleThreads>=0?new HandlerExecutor(name,handleThreads,handler):new HandlerPseudoExecutor(handler);
					SessionWriter filterEncodeProduct=new SessionWriter();
					MessageReceiver filterDecodeProduct=new MessageReceiver(handlerExecutor);
					
					int processorCount=netThreads>0?netThreads:Runtime.getRuntime().availableProcessors()+1;
					processors=new Processor[processorCount];
					for(int i=0;i<processorCount;i++) processors[i]=new Processor(name,filter,filterEncodeProduct,filterDecodeProduct,handlerExecutor,i);
					
					started=true;
				}
			}
		}
	}
	
}
