package com.qq.taf.net;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * 业务逻辑的执行线程池
 * @author fanzhang
 */
public class HandlerExecutor {
	
	private ExecutorService executorService;
	private Handler handler;
	
	private class ExceptionCaught implements Runnable {
		
		private Session session;
		private Throwable exception;
		
		public ExceptionCaught(Session session,Throwable exception) {
			this.session=session;
			this.exception=exception;
		}
		
		public void run() {
			safeCallExceptionCaught(session, exception);
		}
		
	}
	
	private class MessageRecieved implements Runnable {
		
		private Session session;
		private Object message;
		
		public MessageRecieved(Session session,Object message) {
			this.session=session;
			this.message=message;
		}
		
		public void run() {
			try {
				handler.messageRecieved(session, message);
			} catch(Throwable e) {
				safeCallExceptionCaught(session, e);
			}
		}
		
	}
	
	private class SessionClosed implements Runnable {
		
		private Session session;
		
		public SessionClosed(Session session) {
			this.session=session;
		}
		
		public void run() {
			try {
				handler.sessionClosed(session);
			} catch (Throwable e) {
				safeCallExceptionCaught(session, e);
			} finally {
				session.destroy();
			}
		}
		
	}
	
	private class SessionOpened implements Runnable {
		
		private Session session;
		
		public SessionOpened(Session session) {
			this.session=session;
		}
		
		public void run() {
			try {
				handler.sessionOpened(session);
			} catch (Throwable e) {
				safeCallExceptionCaught(session, e);
			}
		}
		
	}
	
	private static class ExecutorThreadFactory implements ThreadFactory {
		
		private AtomicInteger number;
		private String name;
		
		public ExecutorThreadFactory(String name) {
			number=new AtomicInteger(0);
			this.name=name;
		}

		public Thread newThread(Runnable r) {
			return new Thread(r,name+"-Executor-"+number.getAndIncrement());
		}
		
	}
	
	public HandlerExecutor() {
	}
	
	public HandlerExecutor(String name,int threads,Handler handler) {
		ExecutorThreadFactory executorThreadFactory=new ExecutorThreadFactory(name);
		this.executorService=threads>0?Executors.newFixedThreadPool(threads,executorThreadFactory):Executors.newCachedThreadPool(executorThreadFactory);
		this.handler=handler;
	}

	public void exceptionCaught(Session session, Throwable exception) {
		executorService.submit(new ExceptionCaught(session,exception));
	}

	public void messageRecieved(Session session, Object message) {
		executorService.submit(new MessageRecieved(session,message));
	}

	public void sessionClosed(Session session) {
		executorService.submit(new SessionClosed(session));
	}

	public void sessionOpened(Session session) {
		executorService.submit(new SessionOpened(session));
	}
	
	private void safeCallExceptionCaught(Session session,Throwable t) {
		try {
			Logger.log("common caught error",t);
			handler.exceptionCaught(session, t);
		} catch(Throwable tt) {
			Logger.log("handler exceptionCaught error",tt);
		}
	}

}
