package com.qq.taf.net;

import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.Iterator;
import java.util.concurrent.ConcurrentLinkedQueue;

import com.qq.taf.proxy.TafLoggerCenter;
import com.qq.taf.proxy.utils.Millis100TimeProvider;

/**
 * 网络IO处理器
 * @author fanzhang
 */
public class Processor implements Runnable {
	
	protected static final int WRITE_SPIN_COUNT=32;
	protected static final int SELECT_TIMEOUT=100;
	protected static final int IO_TIMEOUT_CHECK_INTERVAL=5000;
	private Filter filter;
	protected SessionWriter filterEncodeProduct;
	protected MessageReceiver filterDecodeProduct;
	protected HandlerExecutor handlerExecutor;
	protected Selector selector;
	protected ConcurrentLinkedQueue<SelectionKey> toCancel;
	protected ConcurrentLinkedQueue<Ternary<Integer, SocketChannel,Integer>> toRegister;
	protected ConcurrentLinkedQueue<SelectionKey> toWrite;
	protected long lastIoTimeoutCheckTime;
	
	public Processor(String name,Filter filter,SessionWriter filterEncodeProduct,MessageReceiver filterDecodeProduct,HandlerExecutor handlerExecutor,int num) {
		try {
			this.filter=filter;
			this.filterEncodeProduct=filterEncodeProduct;
			this.filterDecodeProduct=filterDecodeProduct;
			this.handlerExecutor=handlerExecutor;
			
			selector=Selector.open();
			toCancel=new ConcurrentLinkedQueue<SelectionKey>();
			toRegister=new ConcurrentLinkedQueue<Ternary<Integer,SocketChannel,Integer>>();
			toWrite=new ConcurrentLinkedQueue<SelectionKey>();
			
			new Thread(this,name+"-Processor-"+num).start();
		} catch(Throwable t) {
			Logger.log("processor init error",t);
			Exception.raise(t);
		}
	}

	public void run() {
		try {
			lastIoTimeoutCheckTime=Millis100TimeProvider.INSTANCE.currentTimeMillis();
			
			while(true) {
				int n=selector.select(SELECT_TIMEOUT);
				
				long now=Millis100TimeProvider.INSTANCE.currentTimeMillis();
				Ternary<Integer, SocketChannel,Integer> registerableTernary=null;
				while((registerableTernary=toRegister.poll())!=null) register(registerableTernary,now);
				
				if(n>0) {
					Iterator<SelectionKey> readyKeys=selector.selectedKeys().iterator();
					now=Millis100TimeProvider.INSTANCE.currentTimeMillis();
					SelectionKey readyKey=null;
					while(readyKeys.hasNext()) {
						readyKey=readyKeys.next();
						if(readyKey.isReadable()) read(readyKey,now);
						if(readyKey.isWritable()) ((Session)readyKey.attachment()).scheduleWrite(true);
						readyKeys.remove();
					}
				}
				
				now=Millis100TimeProvider.INSTANCE.currentTimeMillis();
				SelectionKey writableKey=null;
				while((writableKey=toWrite.poll())!=null) write(writableKey,now);
				
				SelectionKey cancelableKey=null;
				while((cancelableKey=toCancel.poll())!=null) cancel(cancelableKey);
				
				now=Millis100TimeProvider.INSTANCE.currentTimeMillis();
				if(now-lastIoTimeoutCheckTime>IO_TIMEOUT_CHECK_INTERVAL) {
					lastIoTimeoutCheckTime=now;
					for(SelectionKey everyKey:selector.keys()) checkIoTimeout(everyKey, now);
				}
			}
		} catch(Throwable t) {
			Logger.log("processor run error",t);
		}
	}
	
	private void checkIoTimeout(SelectionKey key,long now) {
		if(!key.isValid()) return;
		Session session=(Session)key.attachment();
		int ioTimeoutMillis=session.getIoTimeoutMillis();
		if(ioTimeoutMillis<0) return;
		long lastIoTime=Math.max(session.getOpenTime(), Math.max(session.getLastReadTime(), session.getLastWriteTime()));
		if(now-lastIoTime>ioTimeoutMillis) {
			TafLoggerCenter.info("connection io timeout|"+ioTimeoutMillis+"ms|"+session.getLocalAddress()+"|"+session.getRemoteAddress());
			session.close(true);
		}
	}
	
	private void read(SelectionKey key,long now) {
		Session session=(Session)key.attachment();
		try {
			SocketChannel channel=(SocketChannel)key.channel();
			int size=session.getReadBufferSize();
			ByteBuffer buffer=ByteBuffer.allocate(size);
			int read=0;
			int n=0;
			while((n=channel.read(buffer))>0) read+=n;
			if(read>0) {
				buffer.flip();
				filter.decode(session, buffer, filterDecodeProduct);
				if(read==size) session.increaseReadBufferSize();
				else if(read<size>>>1) session.decreaseReadBufferSize();
				session.setLastReadTime(now);
				session.increaseReadBytes(read);
			} else {
				session.decreaseReadBufferSize();
			}
			if(n<0) session.close(false);
		} catch(Throwable t) {
			session.close(true);
			handlerExecutor.exceptionCaught(session, t);
		}
	}
	
	public void scheduleWrite(SelectionKey key) {
		boolean needWakeup=toWrite.isEmpty();
		toWrite.offer(key);
		if(needWakeup) selector.wakeup();
	}
	
	protected void write(SelectionKey key,long now) {	
		if(!key.isValid()) return;
		Session session=(Session)key.attachment();
		try {
			session.scheduleWrite(false);
			setInterestedInWrite(key, false);
			SocketChannel channel=(SocketChannel)key.channel();
			int write=0;
			ConcurrentLinkedQueue<ByteBuffer> queue=session.getWriteQueue();
			ByteBuffer buff=null;
			while((buff=queue.peek())!=null) {
				if(buff==Session.CLOSE_FLAG) {
					queue.clear();
					scheduleCancel(key);
					buff=null;
					break;
				}
				int n=0;
				for(int i=0;i<WRITE_SPIN_COUNT;i++) if((n=channel.write(buff))>0) break;
				if(n==0) break;
				write+=n;
				if(!buff.hasRemaining()) queue.poll();
			}
			if(write>0) {
				session.setLastWriteTime(now);
				session.increaseWriteBytes(write);
			} 
			if(buff!=null) {
				setInterestedInWrite(key, true);
				session.scheduleWrite(true);
			}
		} catch(Throwable t) {
			session.close(true);
			handlerExecutor.exceptionCaught(session, t);
		}
	}
	
	public void scheduleCancel(SelectionKey selectionKey) {
		toCancel.offer(selectionKey);
	}
	
	protected void cancel(SelectionKey selectionKey) {
		try {
			if(!selectionKey.isValid()) return;
			Session session=(Session)selectionKey.attachment();
			selectionKey.cancel();
			selectionKey.attach(null);
			((SocketChannel)selectionKey.channel()).close();
			handlerExecutor.sessionClosed(session);
		} catch(Throwable t) {
			Logger.log("processor cancel error",t);
		}
	}
	
	public void scheduleRegister(SocketChannel socketChannel,int sessionId,int ioTimeoutMillis) {
		toRegister.offer(new Ternary<Integer, SocketChannel,Integer>(sessionId,socketChannel,ioTimeoutMillis));
		selector.wakeup();
	}
	
	protected void register(Ternary<Integer, SocketChannel,Integer> registerTernary,long now) {
		try {
			int sessionId=registerTernary.first;
			SocketChannel socketChannel=registerTernary.second;
			int ioTimeoutMillis=registerTernary.third;
			socketChannel.configureBlocking(false);
			socketChannel.socket().setReuseAddress(true);
			socketChannel.socket().setTcpNoDelay(true);
			socketChannel.socket().setKeepAlive(true);
			SelectionKey key=socketChannel.register(selector, SelectionKey.OP_READ);
			Session session=new Session(sessionId,key,this,filter,filterEncodeProduct,handlerExecutor,ioTimeoutMillis);
			key.attach(session);
			session.setOpenTime(now);
			handlerExecutor.sessionOpened(session);
		} catch (Throwable t) {
			Logger.log("processor register error",t);
		}
	}
	
	protected void setInterestedInWrite(SelectionKey key,boolean interestedInWrite) {
		key.interestOps(interestedInWrite?(SelectionKey.OP_READ | SelectionKey.OP_WRITE):SelectionKey.OP_READ);
	}
	
}
