package com.qq.taf.net.file;

import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicBoolean;

import com.qq.taf.net.HandlerExecutor;
import com.qq.taf.net.Processor;
import com.qq.taf.net.Session;

public class NetSession extends Session {
	private static final int INIT_READ_BUFF_SIZE=1024;
	private static final int MAX_READ_BUFF_SIZE=64*1024;
	private static final int MIN_READ_BUFF_SIZE=8;
	private Map<String,Object> attributes;
	private SelectionKey selectionKey;
	private ConcurrentLinkedQueue<NetDataBuffer> writeQueue;
	private Processor processor;
	private NetFilter filter;
	private NetFilterEncodeProduct filterEncodeProduct;
	private HandlerExecutor handlerExecutor;
	private int readBuffSize;
	private long openTime;
	private long lastReadTime;
	private long lastWriteTime;
	private int readBytes;
	private int writeBytes;
	private int id;
	private AtomicBoolean isScheduledWrite;
	private int ioTimeoutMillis;
	private volatile boolean isOpen;
	
	
	public NetSession(int id,SelectionKey selectionKey,Processor processor,NetFilter filter,NetFilterEncodeProduct filterEncodeProduct,HandlerExecutor handlerExecutor,int ioTimeoutMillis) {
		this.id=id;
		this.selectionKey=selectionKey;
		this.processor=processor;
		this.filter=filter;
		this.filterEncodeProduct=filterEncodeProduct;
		this.handlerExecutor=handlerExecutor;
		attributes=new HashMap<String, Object>();
		writeQueue=new ConcurrentLinkedQueue<NetDataBuffer>();
		readBuffSize=INIT_READ_BUFF_SIZE;
		isScheduledWrite=new AtomicBoolean(false);
		openTime=0;
		lastReadTime=0;
		lastWriteTime=0;
		readBytes=0;
		writeBytes=0;
		this.ioTimeoutMillis=ioTimeoutMillis;
		isOpen=true;
	}
	
	public ConcurrentLinkedQueue<NetDataBuffer> getNetWriteQueue() {
		return writeQueue;
	}
	
	public void setAttribute(String key,Object value) {
		attributes.put(key, value);
	}
	
	public Object getAttribute(String key) {
		return attributes.get(key);
	}
	
	public void removeAttribute(String key) {
		attributes.remove(key);
	}
	
	public void clearAttributes() {
		attributes.clear();
	}
	
	public void write(Object message) {
		try {
			if(isOpen) filter.encode(this, message,filterEncodeProduct);
		} catch(Throwable t) {
			handlerExecutor.exceptionCaught(this, t);
		}
	}
	
	public void close(boolean immediately) {
		isOpen=false;
		if(immediately) {
			writeQueue.clear();
			SelectionKey tmpSelectionKey=selectionKey;
			if(tmpSelectionKey!=null) processor.scheduleCancel(tmpSelectionKey);
		} else {
			writeQueue.offer(NetDataBuffer.CLOSE_DATA);
			scheduleWrite(true);
		}
	}
	
	public long getOpenTime() {
		return openTime;
	}
	
	public long getLastReadTime() {
		return lastReadTime;
	}
	
	public long getLastWriteTime() {
		return lastWriteTime;
	}
	
	public int getReadBytes() {
		return readBytes;
	}
	
	public int getWriteBytes() {
		return writeBytes;
	}
	
	public void setOpenTime(long time) {
		openTime=time;
	}
	
	public void setLastReadTime(long time) {
		lastReadTime=time;
	}
	
	public void setLastWriteTime(long time) {
		lastWriteTime=time;
	}
	
	public void increaseReadBytes(int bytes) {
		readBytes+=bytes;
	}
	
	public void increaseWriteBytes(int bytes) {
		writeBytes+=bytes;
	}
	
	public SocketAddress getRemoteAddress() {
		return ((SocketChannel)selectionKey.channel()).socket().getRemoteSocketAddress();
	}
	
	public SocketAddress getLocalAddress() {
		return ((SocketChannel)selectionKey.channel()).socket().getLocalSocketAddress();
	}
	
	public int getReadBufferSize() {
		return readBuffSize;
	}
	
	public void increaseReadBufferSize() {
		if(readBuffSize<MAX_READ_BUFF_SIZE) readBuffSize=readBuffSize<<1;
	}
	
	public void decreaseReadBufferSize() {
		if(readBuffSize>MIN_READ_BUFF_SIZE) readBuffSize=readBuffSize>>>1;
	}
	
	public int getID() {
		return id;
	}
	
	public void scheduleWrite(boolean isScheduledWrite) {
		if(isScheduledWrite) {
			SelectionKey tmpSelectionKey=selectionKey;
			if(this.isScheduledWrite.compareAndSet(false, true) && tmpSelectionKey!=null) processor.scheduleWrite(tmpSelectionKey);
		} else {
			this.isScheduledWrite.set(false);
		}
	}
	
	public int getIoTimeoutMillis() {
		return ioTimeoutMillis;
	}
	
	public boolean isOpen() {
		return isOpen;
	}
	
	public void destroy() {
		attributes.clear();
		writeQueue.clear();
		selectionKey=null;
	}
	
	@Override
	public String toString() {
		return super.toString()+"("+getLocalAddress()+","+getRemoteAddress()+")";
	}

	@Override
	public final ConcurrentLinkedQueue<ByteBuffer> getWriteQueue() {
		throw new UnsupportedOperationException();
	}
	
	
	public SelectableChannel getChannel() {
		return this.selectionKey.channel();
	}
	
}
