package com.qq.taf.net.udp;

import java.net.SocketAddress;
import java.nio.channels.DatagramChannel;
import java.nio.channels.SelectionKey;
import java.util.HashSet;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

import com.qq.taf.net.Exception;
import com.qq.taf.net.Filter;
import com.qq.taf.net.Handler;
import com.qq.taf.net.HandlerExecutor;
import com.qq.taf.net.HandlerPseudoExecutor;

/**
 * UDP服务端接收器
 * @author easezhang
 *
 */
public class UDPAcceptor {
	private HashSet<SocketAddress> boundAddrs;
	
	private UDPAcceptorProcessor[] processors;
	
	private AtomicInteger nextProcessorIndex;
	
	private AtomicBoolean isStarted;
	
	private int handleThreads;
	
	private Handler handler;
	
	private final String threadName;
	
	/**
	 * UDP服务端接收器
	 * @param threadName 线程名称
	 * @param filter 过滤器
	 * @param handler 处理器
	 * @param netThreads 网络线程个数
	 * @param handleThreads 业务处理线程个数
	 */
	public UDPAcceptor(String threadName, Filter filter, Handler handler, int netThreads, int handleThreads) {
		if(netThreads <= 0) {
			netThreads = Runtime.getRuntime().availableProcessors()+1;
		}
		nextProcessorIndex = new AtomicInteger(0);
		boundAddrs = new HashSet<SocketAddress>();
		isStarted = new AtomicBoolean(false);
		this.handleThreads = handleThreads;
		this.handler = handler;
		this.threadName = threadName;
		processors = new UDPAcceptorProcessor[netThreads];
		for(int i=0;i<processors.length;i++) {
			try {
				processors[i] = new UDPAcceptorProcessor(threadName+"-UDPAcceptorProcessor-"+i, filter, handler);
			}
			catch(Throwable t) {
				Logger.log("create processor error", t);
				processors[i].close();
			}
		}
	}
	
	
	public UDPAcceptor setCallBack(UDPAcceptedCallBack callBack) {
		if(callBack!=null) {
			for(int i=0;i<processors.length;i++) {
				processors[i].setCallBack(callBack);
			}
		}
		return this;
	}


	/**
	 * 启动接收器
	 */
	public void start() {
		if(isStarted.get()) {
			Exception.raise("start error|Acceptor is started, can't be started twice");
		}
		isStarted.set(true);
		HandlerExecutor handlerExecutor = handleThreads>=0?new HandlerExecutor(threadName+"-UDPHandlerExecutor", handleThreads, handler):new HandlerPseudoExecutor(handler);
		for(int i=0;i<processors.length;i++) {
			processors[i].start(handlerExecutor);
		}
	}
	
	public synchronized boolean bind(SocketAddress addr, int timeOut) {
		return bind(addr, timeOut, DefaultConfig.DefaultConfig.getMTU());
	}
	
	/**
	 * 绑定地址，配置channel
	 * @param addr
	 * @throws Throwable
	 */
	public synchronized boolean bind(SocketAddress addr, int timeOut, int receiveBufferSize) {
		if(boundAddrs.contains(addr)) {
			Logger.log("bind address error|"+addr+" is bound!!");
			Exception.raise("bind address error|"+addr+" is bound!!");
		}
		if(isStarted.get()) {
			Logger.log("start error|Acceptor is started, can't bind "+addr+", please bind before start");
			Exception.raise("start error|Acceptor is started, can't bind "+addr+", please bind before start");
		}
		DatagramChannel dc = null;
		SelectionKey key = null;
		try {
			dc = DatagramChannel.open();
			dc.socket().bind(addr);
			Config.DefaultConfig.setReceiveBufferSize(receiveBufferSize);
			Config.DefaultConfig.configDatagramChannel(dc);
			key = processors[Math.abs(nextProcessorIndex.getAndIncrement())%processors.length].regist(dc);
			boundAddrs.add(addr);
			ServerTimeOutGetter.addTimeOut(addr, timeOut);
			return true;
		} catch(Throwable t) {
			Logger.log("bind error|"+addr, t);
			if(key != null) {
				key.cancel();
			}
			if(dc != null) {
				try {
					dc.close();
				} catch(Throwable tt) {
					Logger.log("close error|"+dc,tt);
				}
			}
			Exception.raise(t);
			return false;
		}
	}
}
