package com.qq.taf.net.udp;

import java.io.IOException;
import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.DatagramChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.util.Iterator;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;

import com.qq.taf.net.Filter;
import com.qq.taf.net.MessageReceiver;
import com.qq.taf.net.SessionWriter;
import com.qq.taf.net.Handler;
import com.qq.taf.net.HandlerExecutor;
import com.qq.taf.net.Logger;
import com.qq.taf.proxy.TafLoggerCenter;
import com.qq.taf.proxy.utils.Millis100TimeProvider;

public class UDPAcceptorProcessor extends Thread implements UDPProcessor{
	
	private final static int  SELECT_WAIT_TIME = 100;
	
	private final static int  DEFAULT_SESSION_TIME_OUT = 3000;
	
	private ConcurrentLinkedQueue<UDPSession> flushSessions;
	
	private SessionRecycler sessionRecycler;
	
	private Selector selector;
	
	private SessionWriter filterEncodeProduct;
	
	private MessageReceiver filterDecodeProduct;
	
	private Filter filter;
	
	protected volatile HandlerExecutor handlerExecutor;
	
	private UDPAcceptedCallBack callBack = new UDPAcceptedCallBack();
	
	public UDPAcceptorProcessor(String threadName, Filter filter, Handler handler) throws IOException {
		super(threadName);
		flushSessions = new ConcurrentLinkedQueue<UDPSession>();
		selector = Selector.open();
		this.filterDecodeProduct = new MessageReceiver(this.handlerExecutor);
		this.filterEncodeProduct = new SessionWriter();
		this.filter = filter;
	}
	
	public void start(HandlerExecutor handlerExecutor) {
		this.handlerExecutor = handlerExecutor;
		this.filterDecodeProduct = new MessageReceiver(handlerExecutor);
		//session内部会做线程start，延迟初始化的话就不能放在构造函数做
		sessionRecycler = SessionRecycler.getServerSessioinRecycler();
		start();
	}
	
//	public void addWriteSession(UDPSession ss) {
//		flushSessions.add(ss);
//	}
	
	public void setCallBack(UDPAcceptedCallBack callBack) {
		if(callBack == null) {
			return;
		}
		this.callBack = callBack;
	}

	public SelectionKey regist(DatagramChannel dc) throws Throwable {
		return dc.register(selector, SelectionKey.OP_READ);
	}
	/**
	 * 放入注册队列，唤醒selector
	 * @param addr
	 */
//	public void scheduleRegist(SocketAddress addr) {
//		boolean isEmpty = toRegist.isEmpty();
//		toRegist.add(addr);
//		if(isEmpty) {
//			selector.wakeup();
//		}
//	}
	
	
	public void close() {
		try {
			flushSessions.clear();
			selector.close();
		}
		catch(Throwable t) {
			Logger.log("processor destory error", t);
		}
	}
	
	
	private void read(SelectionKey key) throws Throwable{
		ByteBuffer bb = ByteBuffer.allocate(Config.DefaultConfig.getMTU());
		DatagramChannel dc = (DatagramChannel) key.channel();
		SocketAddress remote = dc.receive(bb);
		if(remote != null) {
			SocketAddress local = dc.socket().getLocalSocketAddress();
			if(!callBack.canAccept(local, remote)) {
				TafLoggerCenter.info("ConnectionIpForbidden " +remote);
				key.cancel();
				return;
			}
			UDPSession ss = sessionRecycler.recycleSession(local, remote);
			if(ss == null) {
				Config.DefaultConfig.configDatagramChannel((DatagramChannel) key.channel());
				ss = new UDPSession(0, remote, local, ServerTimeOutGetter.getTimeOut(local, DEFAULT_SESSION_TIME_OUT), key, filter, filterEncodeProduct, this, handlerExecutor);
				handlerExecutor.sessionOpened(ss);
			}
			sessionRecycler.addSession(local, remote, ss);
			bb.flip();
			ss.setLastAccessTime(Millis100TimeProvider.INSTANCE.currentTimeMillis());
			ss.getFilter().decode(ss, bb, filterDecodeProduct);
		}
	}
	
	private void flushSessions() {
		UDPSession ss = null;
		try {
			while((ss = flushSessions.peek())!=null) {
				ss.setIntrestInWrite(false);
				boolean isWritenAll = write(ss);
				if(!isWritenAll) {
					break;
				}
				else if(!ss.isReadyForWrite() && ss.getWriteQueue().isEmpty()) {
					flushSessions.poll();
				}
			}
		}
		catch(Throwable t) {
			handlerExecutor.exceptionCaught(ss, t);
		}
	}
	
	private boolean write(UDPSession ss) {
		ss.setIntrestInWrite(false);
		ByteBuffer bb = null;
		int sumWriteLen = 0;
		int writeLen = 0;
		boolean retFlag = false;
		try {
			DatagramChannel dc = (DatagramChannel)ss.getSelectionKey().channel();
			while((bb=ss.getWriteQueue().peek())!=null) {
				if(bb == UDPSession.CLOSE_FLAG_DATA) {
					break;
				}
				writeLen = dc.send(bb, ss.getRemoteAddress());
				sumWriteLen += writeLen;
				if (writeLen == 0 || sumWriteLen >= Config.DefaultConfig.getMaxSumSendLen()) {
					ss.setIntrestInWrite(true);
					retFlag =  false;
					break;
				}
				else {//有用吗？
					ss.setIntrestInWrite(false);
				}
				if(writeLen > 0) {
					ss.getWriteQueue().poll();
				}
			}
			if(sumWriteLen>0) {
				ss.setLastAccessTime(Millis100TimeProvider.INSTANCE.currentTimeMillis());
			}
		}
		catch(Throwable t) {
			handlerExecutor.exceptionCaught(ss, t);
		}
		retFlag = true;
		ss.scheduleForReadyWrite(false);
		return retFlag;
	}
	
	public boolean scheduleWrite(UDPSession session) {
		if(session.scheduleForReadyWrite(true)) {
			flushSessions.offer(session);
			selector.wakeup();
			return true;
		}
		else {
			return false;
		}
	}
	
	public void run() {
		while(true) {
			try {
				int eventCount = selector.select(SELECT_WAIT_TIME);
				if(eventCount > 0) {
					Set<SelectionKey> keySet = selector.selectedKeys();
					Iterator<SelectionKey> itor = keySet.iterator();
					while(itor.hasNext()) {
						SelectionKey key = itor.next();
						if(key.isReadable()) {
							read(key);
							itor.remove();
						}
					}
				}
				flushSessions();//此处如果没有写完，如何再次环形selector？只能采用有限的select等待？
			}
			catch(Throwable t) {
				Logger.log("processor runnnig error ",t);
			}
		}
	}

	@Override
	public int scheduleConnect(DatagramChannel dc, SocketAddress remote,
			int sessionTimeOut, int sessionId) {
		throw new UnsupportedOperationException("UDPAcceptor can't scheduleConnect");
	}

	@Override
	public void scheduleCancel(UDPSession session) {
		//donothing 服务端调用该方法不能把绑定的端口也关了，因此此处无
	}
}
