package com.qq.taf.net.udp;

import java.net.SocketAddress;
import java.nio.channels.DatagramChannel;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

import com.qq.taf.net.Exception;
import com.qq.taf.net.Filter;
import com.qq.taf.net.SessionWriter;
import com.qq.taf.net.Handler;
import com.qq.taf.net.HandlerExecutor;
import com.qq.taf.net.HandlerPseudoExecutor;
import com.qq.taf.net.SessionIdGenerator;

/**
 * UDP连接器
 * @author easezhang
 *
 */
public class UDPConnector {
	
	private UDPConnectorProcessor[] processors;
	
	private AtomicInteger nextProcessorIndex;
	
	private AtomicBoolean isStarted;
	
	private int handleThreads;
	
	private Handler handler;
	
	private final String threadName;
	
	private HandlerExecutor handlerExecutor;
	
	private SessionWriter filterEncodeProduct;
	
	private SessionIdGenerator sessionIdGenerator;
	/**
	 * UDP连接器构造函数
	 * @param threadName 线程名称
	 * @param filter 过滤器
	 * @param handler 处理器
	 * @param netThreads 网络线程数
	 * @param handleThreads 业务处理线程数
	 */
	public UDPConnector(String threadName, Filter filter, Handler handler, int netThreads, int handleThreads) {
		this.threadName = threadName;
		if(netThreads <= 0) {
			netThreads = Runtime.getRuntime().availableProcessors();
		}
		nextProcessorIndex = new AtomicInteger(0);
		isStarted = new AtomicBoolean(false);
		this.handleThreads = handleThreads;
		this.handler = handler;
		processors = new UDPConnectorProcessor[netThreads];
		filterEncodeProduct = new SessionWriter();
		sessionIdGenerator=new SessionIdGenerator();
		for(int i=0;i<processors.length;i++) {
			try {
				processors[i] = new UDPConnectorProcessor(threadName+"-UDPConnectorProcessor-"+i, filter, handler, filterEncodeProduct);
			}
			catch(Throwable t) {
				Logger.log("create processor error", t);
				processors[i].close();
			}
		}
	}
	
	/**
	 * <p>注入一个SessionId生成器
	 * <p>说明：仅当多个Connector共用一套Filter与Handler，并且需要使用connect方法返回的SessionId的情况下，必须同时注入一个共用的SessionIdGenerator实例以确保SessionId被生成时的唯一性。
	 * @param sessionIdGenerator
	 */
	public void setSessionIdGenerator(SessionIdGenerator sessionIdGenerator) {
		this.sessionIdGenerator=sessionIdGenerator;
	}
	
	/**
	 * 启动连接器
	 */
	private void start() {
		if(isStarted.get()) {
			return;
		}
		synchronized(isStarted) {
			if(isStarted.get()) {
				return;
			}
			handlerExecutor = handleThreads>=0?new HandlerExecutor(threadName+"-UDPHandlerExecutor", handleThreads, handler):new HandlerPseudoExecutor(handler);
			for(int i=0;i<processors.length;i++) {
				processors[i].start(handlerExecutor);
			}
			isStarted.set(true);
		}
	}
	
	private DatagramChannel newDatagramChannel() throws Throwable {
		DatagramChannel dc = DatagramChannel.open();
		Config.DefaultConfig.configDatagramChannel(dc);
		return dc;
	}
	
	/**
	 * 连接远程服务端口，调用前务必先调用start()
	 * @param remote 远程地址
	 * @param sessionTimeOut session超时时间
	 * @return UDPSessionID
	 */
	public int connect(SocketAddress remote, int sessionTimeOut) {
		int sessionId=0;
		try {
			start();
			DatagramChannel dc = newDatagramChannel();
			//dc.socket().setSoTimeout(sessionTimeOut);//非阻塞模式无效
			dc.connect(remote);
			UDPProcessor processor = processors[(nextProcessorIndex.getAndIncrement()&1023)%processors.length];
			sessionId = sessionIdGenerator.newSessionId();
			processor.scheduleConnect(dc, remote, sessionTimeOut, sessionId);
		} catch(Throwable t) {
			Logger.log("connect error|"+remote,t);
			Exception.raise(t);
		}
		return sessionId;
	}
	
}
