package com.qq.taf.net.udp;

import java.io.IOException;
import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.DatagramChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.util.Iterator;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.LinkedBlockingQueue;

import com.qq.taf.net.Filter;
import com.qq.taf.net.MessageReceiver;
import com.qq.taf.net.SessionWriter;
import com.qq.taf.net.Handler;
import com.qq.taf.net.HandlerExecutor;
import com.qq.taf.proxy.utils.Millis100TimeProvider;


public class UDPConnectorProcessor extends Thread implements UDPProcessor{
	
	private final static int  SELECT_WAIT_TIME = 100;
	
	private ConcurrentLinkedQueue<UDPSession> flushSessions;
	
	private ConcurrentLinkedQueue<Quin<DatagramChannel, SocketAddress, SocketAddress, Integer, Integer>> toRegist;
	
	private ConcurrentLinkedQueue<UDPSession> toCancel;
	
	private Selector selector;
	
	private SessionWriter filterEncodeProduct;
	
	private MessageReceiver filterDecodeProduct;
	
	private Filter filter;
	
	protected volatile HandlerExecutor handlerExecutor;
	
	private final ChannelIniter[] channelIniters;
	
	private int INITER_COUNT = 1;
	
	private int nextIniterIndex = 0;
	
	//private final int DEFAULT_MAX_OPEN_CONNECTION = 16;
	
	public UDPConnectorProcessor(String threadName, Filter filter, Handler handler, SessionWriter filterEncodeProduct) throws IOException {
		super(threadName);
		channelIniters = new ChannelIniter[INITER_COUNT];
		for(int i=0;i<channelIniters.length;i++) {
			channelIniters[i] = new ChannelIniter(threadName+"-"+i);
		}
		try {
			flushSessions = new ConcurrentLinkedQueue<UDPSession>();
			toCancel = new ConcurrentLinkedQueue<UDPSession>();
			toRegist = new ConcurrentLinkedQueue<Quin<DatagramChannel, SocketAddress, SocketAddress, Integer, Integer>>();
			selector = Selector.open();
			this.filterEncodeProduct = filterEncodeProduct; 
			this.filter = filter;
			handlerExecutor = null;
		}
		catch(Throwable t) {
			Logger.log("create UDPConnectorProcessor error", t);
			close();
		}
	}
	
	private class ChannelIniter extends Thread {
		
		private LinkedBlockingQueue<UDPSession> toClose;
		
		public ChannelIniter(String namePrefix) {
			super(namePrefix+"-ChannelIniter");
			toClose = new LinkedBlockingQueue<UDPSession>();
		}
		
		public void run() {
			UDPSession ss = null;
			while(true) {
				try {
					while((ss = toClose.poll())!= null) {
						closeSession(ss);
					}
					Thread.sleep(SELECT_WAIT_TIME);
				}
				catch(InterruptedException e) {
					//do null就是为了唤醒
				}
				catch(Throwable t) {
					Logger.log("ChannelIniter run error", t);
				}
			}
		}
		
		
		public void addClosingSession(UDPSession ss) {
			boolean isEmpty = toClose.isEmpty();
			toClose.offer(ss);
			if(isEmpty) {
				interrupt();
			}
		}
		
		private void closeSession(UDPSession ss) {
			try {
				ss.close(true);
				ss.getSelectionKey().attach(null);
				ss.getSelectionKey().channel().close();
			}
			catch(Throwable t) {
				Logger.log("close session error", t);
			}
		}
	}
	
	public void scheduleCancel(UDPSession ss) {
		toCancel.offer(ss);
		//selector.wakeup();此处让关闭动作随着其他wakeup完成即可
	}
	
	private void cancel() {
		UDPSession ss = null;
		while((ss=toCancel.poll())!= null) {
			doCancel(ss);
		}
	}
	
	private void doCancel(UDPSession ss) {
		try {
			ss.getSelectionKey().cancel();
			channelIniters[((++nextIniterIndex)&(1024-1))%channelIniters.length].addClosingSession(ss);
		}
		catch(Throwable t) {
			Logger.log("close session error", t);
		}
	}
	
	private void regist() {
		Quin<DatagramChannel, SocketAddress, SocketAddress, Integer, Integer> ty = null;
		while((ty=toRegist.poll())!= null) {
			doConnect(ty.first, ty.second, ty.third, ty.forth, ty.fifth);
		}
	}
	
		
	private void doConnect(DatagramChannel dc, SocketAddress remote, SocketAddress local, int sessionTimeout, int sessionId) {
		try {
			SelectionKey key = dc.register(selector, SelectionKey.OP_READ);
			UDPSession ss = new UDPSession(sessionId, remote, local, sessionTimeout, key, filter, filterEncodeProduct, this, handlerExecutor);
			key.attach(ss);
			handlerExecutor.sessionOpened(ss);
		}
		catch(Throwable t) {
			try {
				dc.close();
				Logger.log("connect "+remote+" error", t);
			}
			catch(Throwable tt) {
				Logger.log("connect "+remote+" error", tt);
			}
		}
	}
		
	
	
	public void start(HandlerExecutor handlerExecutor) {
		this.handlerExecutor = handlerExecutor;
		this.filterDecodeProduct = new MessageReceiver(this.handlerExecutor);
		for(int i=0;i<channelIniters.length;i++) {
			channelIniters[i].start();
		}
		super.start();
	}
	
	public void close() {
		try {
			flushSessions.clear();
			toRegist.clear();
			selector.close();
		}
		catch(Throwable t) {
			Logger.log("processor destory error", t);
		}
	}
	
	private void flushSessions() {
		final UDPSession firstSession = flushSessions.peek();
		UDPSession ss = null;
		int writeLen = 0;
		int totalWriteLen = 0;
		boolean isFirst = true;
		try {
			long now = Millis100TimeProvider.INSTANCE.currentTimeMillis();
			
			while((ss = flushSessions.poll())!=null) {
				ss.scheduleForReadyWrite(false);
				if(ss == firstSession && !isFirst) {
					//循环一周
					isFirst =true;
					break;
				}
				ss.setIntrestInWrite(false);
				if(!ss.isOpen()) {
					scheduleCancel(ss);
				}
				writeLen = write(ss, now);
				if(isFirst) {
					isFirst = false;
				}
				if(writeLen < 0) {
					continue;
				}
				totalWriteLen += writeLen;
				if(writeLen==0) {
					//未写成功或者没写完
					ss.setIntrestInWrite(true);
					break;
				}
				if(totalWriteLen >= Config.DefaultConfig.getMaxSumSendLen()) {
					//一次发送字节太多
					ss.setIntrestInWrite(true);
					break;
				}
			}
		}
		catch(Throwable t) {
			handlerExecutor.exceptionCaught(ss, t);
		}
	}
	
	/**
	 * 发送一次Session数据
	 * @param ss Session
	 * @return 发送的字节数
	 */
	private int write(UDPSession ss, long now) {
		int writeLen = 0;
		try {
			ss.setIntrestInWrite(false);
			ConcurrentLinkedQueue<ByteBuffer> writeData = ss.getWriteQueue();
			DatagramChannel dc = (DatagramChannel) ss.getSelectionKey().channel();
			ByteBuffer bb = null;
			bb = writeData.peek();
			if(bb == UDPSession.CLOSE_FLAG_DATA) {
				scheduleCancel(ss);
				return -1;
			}
			if(bb != null) {
				//int oldLimit = bb.limit();
				//bb.limit(Math.min(oldLimit, bb.position()+Config.DefaultConfig.getMTU()));
				writeLen = dc.write(bb);
				if(writeLen > 0) {
					writeData.poll();
				}
				if(writeLen == 0) {
					ss.setIntrestInWrite(true);
				}
				else {
					ss.setLastAccessTime(now);
				}
			}
		}
		catch(Throwable t) {
			ss.close(true);
			this.handlerExecutor.exceptionCaught(ss, t);
		}
		if(!ss.getWriteQueue().isEmpty()) {
			ss.setIntrestInWrite(true);
		}
		return writeLen;
	}
	
	private void read(SelectionKey key, long now) {
		UDPSession ss = null;
		try {
			DatagramChannel dc =(DatagramChannel)key.channel();
			ss = (UDPSession)key.attachment();
			if(ss == null) {
				throw new Exception("key is not attach session");
			}
			ByteBuffer bb = ByteBuffer.allocate(Config.DefaultConfig.getMTU());
			int readLen = dc.read(bb);
			if(readLen < 0) {
				Logger.log("read "+readLen);
				ss.close(false);
			}
			if(readLen > 0) {
				bb.flip();
				filter.decode(ss, bb, filterDecodeProduct);
				ss.setLastAccessTime(now);
			}
		}
		catch(Throwable t) {
			if(ss != null) {
				ss.close(true);
				handlerExecutor.exceptionCaught(ss, t);
			}
		}
	}
	
	public void run() {
		long now = 0;
		while(true) {
			try {
				int eventCount = selector.select(SELECT_WAIT_TIME);
				regist();
				if(eventCount > 0) {
					Iterator<SelectionKey> itor = selector.selectedKeys().iterator();
					now = Millis100TimeProvider.INSTANCE.currentTimeMillis();
					while(itor.hasNext()) {
						SelectionKey key = itor.next();
						if(key.isValid()&&key.isReadable()) {
							read(key, now);
						}
						if(key.isValid()&&key.isWritable()) {
							scheduleWrite((UDPSession) key.attachment());
						}
						itor.remove();
					}
				}
				flushSessions();
				cancel();
			}
			catch(Throwable t) {
				Logger.log("UDPConnector run error", t);
			}
		}
	}

	@Override
	public boolean scheduleWrite(UDPSession session) {
		if(session.scheduleForReadyWrite(true)) {
			flushSessions.offer(session);
			selector.wakeup();
			return true;
		}
		else {
			return false;
		}
	}

	@Override
	public int scheduleConnect(DatagramChannel dc, SocketAddress remote,
			int sessionTimeOut, int sessionId)  {
		//channelIniters[((++nextIniterIndex)&(1024-1))%channelIniters.length].addNewChannel(remote, sessionTimeOut);
		boolean isEmpty = toRegist.isEmpty();
		toRegist.offer(new Quin<DatagramChannel, SocketAddress, SocketAddress, Integer, Integer>(dc, remote, dc.socket().getLocalSocketAddress(), sessionTimeOut, sessionId));
		if(isEmpty) {
			selector.wakeup();
		}
		return 0;
	}
}
