package com.qq.taf.net.udp;

import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.util.LinkedHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicBoolean;

import com.qq.taf.net.Filter;
import com.qq.taf.net.SessionWriter;
import com.qq.taf.net.HandlerExecutor;
import com.qq.taf.net.Session;
import com.qq.taf.proxy.utils.Millis100TimeProvider;

public class UDPSession extends Session implements Expirable {

	public final static ByteBuffer CLOSE_FLAG_DATA = ByteBuffer.allocate(0);

	private final int id;
	
	// 上次访问时间
	private long lastAccessTime;

	// Session属性
	private final LinkedHashMap<String, Object> attrs;

	// 是否已加入写队列，防止重复加入
	private AtomicBoolean isReadyWrite;

	// session超时时间
	private final int timeOut;

	// 写数据队列
	private final ConcurrentLinkedQueue<ByteBuffer> writeQueue;

	// selectorkey
	private volatile SelectionKey selectionKey;

	// 远程地址
	private final SocketAddress remoteAddress;

	// 本地地址
	private final SocketAddress localAddress;

	private SessionWriter filterEncodeProduct;

	private volatile boolean isClosed;

	private Filter filter;

	private UDPProcessor processor;

	private HandlerExecutor handlerExecutor;

	/**
	 * UDP回话
	 * @param remoteAddress 远程地址
	 * @param localAddress 本地地址
	 * @param timeOut session超时时间
	 * @param selectionKey selectioinKey
	 * @param filter 
	 * @param filterEncodeProduct
	 * @param processor
	 * @param handlerExecutor
	 */
	public UDPSession(int sessionId, SocketAddress remoteAddress, SocketAddress localAddress,
			int timeOut, SelectionKey selectionKey, Filter filter,
			SessionWriter filterEncodeProduct,
			UDPProcessor processor, HandlerExecutor handlerExecutor) {
		id = sessionId;
		long now = Millis100TimeProvider.INSTANCE.currentTimeMillis();
		this.lastAccessTime = now;
		attrs = new LinkedHashMap<String, Object>();
		isReadyWrite = new AtomicBoolean(false);
		writeQueue = new ConcurrentLinkedQueue<ByteBuffer>();
		this.timeOut = timeOut;
		this.remoteAddress = remoteAddress;
		this.localAddress = localAddress;
		this.selectionKey = selectionKey;
		this.filter = filter;
		isClosed = false;
		this.filterEncodeProduct = filterEncodeProduct;
		this.processor = processor;
		this.handlerExecutor = handlerExecutor;
	}

	
	public int getID() {
		return id;
	}


	public void setKey(SelectionKey key) {
		this.selectionKey = key;
	}

	public boolean isOpen() {
		return !isClosed;
	}

	public void setLastAccessTime(long time) {
		this.lastAccessTime = time;
	}
	
	@Override
	public long getLastReadTime() {
		return lastAccessTime;
	}
	
	@Override
	public long getLastWriteTime() {
		return lastAccessTime;
	}

	public SelectionKey getSelectionKey() {
		return this.selectionKey;
	}

	@Override
	public boolean isExpired() {
		long now = Millis100TimeProvider.INSTANCE.currentTimeMillis();
		//boolean isEmpty = writeQueue.isEmpty();
		//此处判断标准是否需要看写队列为空？
		if (now - this.lastAccessTime > timeOut ||isClosed) {
			return true;
		}
		return false;
	}

	public SocketAddress getRemoteAddress() {
		return remoteAddress;
	}

	public SocketAddress getLocalAddress() {
		return localAddress;
	}

	public void setAttribute(String attrName, Object value) {
		attrs.put(attrName, value);
	}

	public Object getAttribute(String attrName) {
		return attrs.get(attrName);
	}

	public void removeAttribute(String attrName) {
		attrs.remove(attrName);
	}

	public void clearAttributes() {
		this.attrs.clear();
	}

	/**
	 * 设置是否已准备好可写
	 * 
	 * @param flag
	 * @return
	 */
	public boolean scheduleForReadyWrite(boolean flag) {
		if (flag) {
			return this.isReadyWrite.compareAndSet(false, true);
		} else {
			this.isReadyWrite.set(false);
			return true;
		}
	}

	public boolean isReadyForWrite() {
		return this.isReadyWrite.get();
	}

	public ConcurrentLinkedQueue<ByteBuffer> getWriteQueue() {
		return this.writeQueue;
	}

	public void write(Object obj) {
		try {
			if (isClosed) {
				throw new Exception("session is closed");
			}
			filter.encode(this, obj, filterEncodeProduct);
		} catch (Throwable e) {
			handlerExecutor.exceptionCaught(this, e);
		}
	}

	public void destroy() {
		try {
			//this.attrs.clear();
			this.writeQueue.clear();
		} catch (Throwable t) {
			Logger.log("session destroy error", t);
		}
	}

	public void close(boolean isImmedately) {
		if(isClosed) {
			return;
		}
		isClosed = true;
		handlerExecutor.sessionClosed(this);
		if (isImmedately) {
			destroy();
			//对于client要销毁持有资源
			processor.scheduleCancel(this);
		} else {
			this.writeQueue.offer(CLOSE_FLAG_DATA);
			processor.scheduleWrite(this);
		}
	}

	public void setIntrestInWrite(boolean isIntrestInWrite) {
		if (!this.selectionKey.isValid()) {
			return;
		}
		if (isIntrestInWrite) {
			this.selectionKey.interestOps(this.selectionKey.interestOps() | SelectionKey.OP_WRITE);
		} else {
			this.selectionKey.interestOps(this.selectionKey.interestOps() & (~SelectionKey.OP_WRITE));
		}
	}

	public Filter getFilter() {
		return filter;
	}
	
	public void scheduleWrite(boolean isScheduledWrite) {
		processor.scheduleWrite(this);
	}

	@Override
	public String toString() {
		return getClass().getName()+"@"+Integer.toHexString(hashCode())+"("+getLocalAddress()+","+getRemoteAddress()+")";
	}
	
}
