package com.qq.taf.proxy.conn;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;

import com.qq.taf.ResponsePacket;
import com.qq.taf.StatMicMsgHead;
import com.qq.taf.StatSampleMsg;
import com.qq.taf.cnst.JCEASYNCCALLTIMEOUT;
import com.qq.taf.cnst.JCEPROXYCONNECTERR;
import com.qq.taf.cnst.JCEVERSION;
import com.qq.taf.proxy.ProxyStatManager;
import com.qq.taf.proxy.PrxConfig;
import com.qq.taf.proxy.SendMessage;
import com.qq.taf.proxy.SendMsgStatBody;
import com.qq.taf.proxy.ServantFuture;
import com.qq.taf.proxy.ServantProxy;
import com.qq.taf.proxy.ServiceInfos;
import com.qq.taf.proxy.ServiceInfosRefresher;
import com.qq.taf.proxy.TAdapterSelector;
import com.qq.taf.proxy.TafLoggerCenter;
import com.qq.taf.proxy.codec.JceMessage;
import com.qq.taf.proxy.exec.TafCallTimeoutException;
import com.qq.taf.proxy.exec.TafException;
import com.qq.taf.proxy.exec.TafProxyNoConnException;
import com.qq.taf.proxy.utils.TafUtils;
import com.qq.taf.proxy.utils.TimeoutHandler;

/**
 * 维持了一个ServantProxy到它的servant的所有连接 负责刷新 新建连接
 * 
 * @author albertzhu
 * 
 */
public class ServantConnManager {

	// proxy名
	String objectName;
	// proxy模块名称
	String processName;
	// 网络连接接口 根据请求打开连接并转发IoSession事件
	EndPointAdapter adapter;
	// 信息序列号
	static AtomicInteger seq = new AtomicInteger();
	ServantProxy proxy;
	// 服务器选择器
	TAdapterSelector selector = new TAdapterSelector();

	// 隶属的proxy配置文件
	PrxConfig proxyConfig;

	volatile ServiceInfos serviceInfos;

	//以 协议>主机ip:端口 作为主键 TODO 性能优化
	ConcurrentHashMap<String, ServantEndPoint> aliveServiceMap = new ConcurrentHashMap<String, ServantEndPoint>();

	private Boolean isInit = new Boolean(false);

	FutureTimeoutHandler timeoutHandler;
	
	/**
	 * 
	 * @param processName
	 *            proxy模块名称
	 * @param proxyConfig
	 *            隶属的proxy配置文件
	 * @param adapter
	 *            网络连接接口 根据请求打开连接并转发IoSession事件
	 */
	public ServantConnManager(String processName, PrxConfig proxyConfig, EndPointAdapter adapter) {
		this.processName = processName;
		this.objectName = proxyConfig.getObjectName();
		this.adapter = adapter;
		this.serviceInfos = proxyConfig.serviceInfos;
		this.proxyConfig = proxyConfig;
		timeoutHandler = new FutureTimeoutHandler();
	}
	
	public String toString() {
		return objectName+"'s Router(ID="+proxy.getID()+")";
	}
	/**
	 * 初始化配置 将配置中的服务器地址信息加载 建立连接 线程安全
	 */
	public void initConfigs() {
		ServiceInfos tmpInfos = null;
		if(!isInit) {
			synchronized(isInit) {
				if(!isInit) {
					if(serviceInfos.getActiveServicesSize() == 0) {
						TafLoggerCenter.info(this+" initConfigs from registry");
						try {
							tmpInfos = ServiceInfosRefresher.INSTANCE.getServiceInfos(proxy);
							if (tmpInfos != null && tmpInfos.getActiveServicesSize() > 0) {
								refreshEndPoints(tmpInfos);
							}
							else {
								refreshEndPoints(serviceInfos);
							}
						}
						catch(Throwable th) {
							TafLoggerCenter.error(this+" init from registry error, and initConfigs from "+serviceInfos);
							refreshEndPoints(serviceInfos);
						}
					}
					else {
						TafLoggerCenter.info(objectName+" initConfigs from "+serviceInfos);
						refreshEndPoints(serviceInfos);
					}
					isInit = true;
				}
			}
		}
	}
	
	/**
	 *更新从register中心取得service列表
	 * 
	 * @param newServiceInfos
	 * @threadSafe
	 */
	public void refreshEndPoints(ServiceInfos newServiceInfos) {
		HashSet<String> activeServiceKeys = newServiceInfos.getActiveServiceKeys();
		TafLoggerCenter.debug(this + " " + objectName + " refreshEndPoints start, activeServiceKeys.size():" + activeServiceKeys.size());
		// 移除不在新下发的service列表中的服务器
		for (Entry<String, ServantEndPoint> entry : aliveServiceMap.entrySet()) {
			String key = entry.getKey();			

			if (!activeServiceKeys.contains(key)) {
				try {
					removeService(key);
					TafLoggerCenter.info(this+" removeNotexistService " + key);
				} catch (Exception e) {
					TafLoggerCenter.error(this+" removeNotexistService " + key + " error " + e, e);
				}
			}
		}
		// 移除下发列表中标记失败的service
		List<EndPointInfo> inactiveServices = newServiceInfos.getInActiveServices();
		for (EndPointInfo serviceInfo : inactiveServices) {
			try {
				removeService(serviceInfo.getKey());
				TafLoggerCenter.info(this+ " refresh removeInactiveService " + serviceInfo.getKey());
			} catch (Exception e) {
				TafLoggerCenter.error(this+" refresh removeInactiveService " + serviceInfo + " error " + e, e);
			}
		}
		
		List<EndPointInfo> activeServices = newServiceInfos.getActiveServices();
		for (EndPointInfo si : activeServices) {
			TafLoggerCenter.info("check "+proxyConfig.getObjectName()+" service "+si);
			if (si.active == false && si.protocol.equals("udp")) {
				TafLoggerCenter.info("not check failed udp server "+si);
			} else {
				if(!aliveServiceMap.containsKey(si.getKey())) {
					//如果是活的，还要刷新到活的endpoint map里
					TafLoggerCenter.info(this+" add "+si+" to aliveServiceMap");
					ServantEndPoint tmpPoint = new ServantEndPoint(this, si, timeoutHandler);
					aliveServiceMap.putIfAbsent(si.getKey(), tmpPoint);
				}
			}
		}
		
		this.serviceInfos = newServiceInfos;
		TafLoggerCenter.debug(this+" refreshEndPoints end, serviceInfos="+serviceInfos );
	}

	/**
	 * 从存活service列表中移除一个service
	 * 外面调用时会加锁
	 * @param service
	 */
	private void removeService(String infoKey) {
		ServantEndPoint point = aliveServiceMap.remove(infoKey);
		if (point!=null) {
			point.closeAll();
		}
	}


	public void closeAllConns() {
		for (Entry<String, ServantEndPoint> entry : aliveServiceMap.entrySet()) {
			try {
				ServantEndPoint value = entry.getValue();
				value.closeAll();
			} catch (Exception e) {
				TafLoggerCenter.error(this+" closeAllConns error", e);
			}
		}
		aliveServiceMap.clear();
	}

	/**
	 * 选择一个合适的服务器发送信息
	 * 
	 * @param msg
	 * @threadSafe
	 */
	public void sendMsg(SendMessage msg) {
		ServantEndPoint endPoint = getAliveEndPoint(msg);
		endPoint.sendMsg(msg);
		TafLoggerCenter.debug(proxyConfig.getObjectName() + " " + endPoint.getInfo() + " send msg "
				+ msg.getFuture().getSeq());
	}

	/**
	 * 再次尝试发送 直接将异常通知等待的future
	 * 
	 * @param msg
	 */
	public void tryAgainSend(SendMessage msg) {
		boolean failed = true;
		try {
			getAliveEndPoint(msg).sendMsg(msg);
			failed = false ;
		} catch (Exception ex) {
			TafLoggerCenter.info(this+" tryAgainSend msg Exception " + ex, ex);
			msg.getFuture().setFailure(new TafException("second send " + ex.toString(), ex));
		} catch (Throwable ex) {
			TafLoggerCenter.info(this+" tryAgainSend msg Throwable " + ex, ex);
			msg.getFuture().setFailure(new TafException(ex.toString(), ex));
		} finally {
			if (failed && msg.getFuture().isAsync()) {
				JceMessage jceMsg = JceMessage.createErrorRespMessage(JCEPROXYCONNECTERR.value, msg.getFuture()
						.getSeq());
				jceMsg.setFuture(msg.getFuture());
				proxyConfig.getTafThreadPool().receiveMsg(jceMsg);
			}
		}
	}

	/**
	 * 创建一个失败的异步回应并激活callBackHandler
	 * 
	 * @param future
	 */
	void createErrRespDispatch(ServantFuture future, int iret) {
		String methodName = future.getSFuncName();
		ResponsePacket rs = new ResponsePacket((short) JCEVERSION.value, (byte) 0, future.getSeq(), 0, iret,
				new byte[] {}, new HashMap<String, String>(), "");
		future.getHandler()._onDispatch(methodName, rs);
	}

	/**
	 * 创建一个msg future 并设置序列号、允许等待时间和异步调用标示
	 * 
	 * @param allowWaiteTime
	 * @param isAsync
	 * @return
	 */
	public ServantFuture createJceFuture(long allowWaiteTime, boolean isAsync) {
		ServantFuture future = new ServantFuture(getNextSeq(), allowWaiteTime);
		future.setAsync(isAsync);
		return future;
	}

	/**
	 * 序列号生成器
	 * 
	 * @return
	 */
	protected int getNextSeq() {
		return seq.incrementAndGet();
	}


	Object waitConnLock = new Object();

	// /////////////////////////////////////////////////////////////////////////////

	AtomicLong lastConnTime = new AtomicLong(0);

	private ServantEndPoint getAliveEndPoint(SendMessage msg) {
		
		ServantEndPoint ret = null;
		int maxTryCount = serviceInfos.getActiveServicesSize();
		int i = 0;
		EndPointInfo endPointInfo = null ;
		while (i < maxTryCount) {
			endPointInfo = selector.getAliveConn(serviceInfos, msg);
			if (null != endPointInfo) {
				//超时策略检查
				boolean canConnect = SwitchInfoFactory.canConnect(endPointInfo, this);				
				if(canConnect) {
					ret = aliveServiceMap.get(endPointInfo.getKey());
					if (ret == null) {
						// aliveServiceMap和活着的endPointInfo是同步的,不应该出现为null
						TafLoggerCenter.error(this + " getAliveEndPoint can not find this service " + endPointInfo);
						ret = new ServantEndPoint(this, endPointInfo, timeoutHandler);
						aliveServiceMap.putIfAbsent(endPointInfo.getKey(), ret);
					}
					else {
						break;
					}
				}
			} else {
				TafLoggerCenter.info(this+" failed getAliveConn for " + objectName);
			}
			i++;
		}
		
		i = 0;
		if(ret == null) { //都超时了，按照选取策略选取一个
			while (i < maxTryCount) {
				endPointInfo = selector.getAliveConn(serviceInfos, msg);
				if (null != endPointInfo) {
					//超时策略检查
					ret = aliveServiceMap.get(endPointInfo.getKey());
					if (ret == null) {
						// aliveServiceMap和活着的endPointInfo是同步的,不应该出现为null
						TafLoggerCenter.error(this + " getAliveEndPoint2 can not find this service " + endPointInfo);
						ret = new ServantEndPoint(this, endPointInfo, timeoutHandler);
						aliveServiceMap.putIfAbsent(endPointInfo.getKey(), ret);
					}
					else {
						break;
					}
				} else {
					TafLoggerCenter.info(this+" failed getAliveConn for " + objectName);
				}
				i++;
			}			
		}
		if (null == ret) {
			throw new TafProxyNoConnException("can not find this service " + endPointInfo + " " + aliveServiceMap + " "
					+ objectName);
		}
		return ret;

	}

	// ///////////////////////////////////////////////////////////////////////////////////////

	/**
	 * 设置统计区间
	 * 
	 * @param timeStatInterv
	 */
	public void setTimeStatInterv(List<Integer> timeStatInterv) {
		ProxyStatManager.getInstance().setPointStatInterv(proxy);
	}

	public void setServantProxy(ServantProxy proxy) {
		this.proxy = proxy;
	}

	public ServantProxy getServantProxy() {
		return proxy;
	}
	
	public List<Integer> getTimeStatInterv() {
		return ProxyStatManager.getInstance().getStatTool(proxy).getStatIntervals();
	}

	public TAdapterSelector getSelector() {
		return selector;
	}

	public void setSelector(TAdapterSelector selector) {
		this.selector = selector;
	}

	class FutureTimeoutHandler implements TimeoutHandler<Integer, ServantFuture> {
		public void timeout(Integer k, ServantFuture future, long now) {
			if (future.isAsync()) {
				// 没有线程在等待回应
				future.setFailure(new TafCallTimeoutException("asyncMsg " + future.getSeq() + " waiteResp timeout "));
				JceMessage jceMsg = JceMessage.createErrorRespMessage(JCEASYNCCALLTIMEOUT.value, future.getSeq());
				jceMsg.setFuture(future);
				proxyConfig.getTafThreadPool().receiveMsg(jceMsg);
			} else {
				future.setFailure(new TafCallTimeoutException("msg " + future.getSeq() + " waiteResp timeout "));
			}
			String key = future.getEndPointKey();
			ServantEndPoint endPoint = aliveServiceMap.get(key);
			if (null != endPoint) {
				future.onNetCallFinished();
				StatMicMsgHead head = TafUtils.getHead(processName, future.getSServantName(), future.getSFuncName(), "",
						future.getRemoteHost(), future.getRemotePort(), 0, proxyConfig.getSetName() , proxyConfig.getSetArea(), proxyConfig.getSetId(), endPoint.getInfo().getSetDivision());
				endPoint.onCallFinished(head, SendMsgStatBody.CallStatusTimeout, future.getCostTime(), "", false);
			} else {
				TafLoggerCenter.info(this+" timeoutHandler endPoint "+key+" inactive");
			}
		}
	}

	public ConcurrentHashMap<StatMicMsgHead, SendMsgStatBody> getStatmsg() {
		return ProxyStatManager.getInstance().getStatTool(proxy).getStats();
	}
	
	public ConcurrentLinkedQueue<StatSampleMsg> StatSampleMsgQueue = new ConcurrentLinkedQueue<StatSampleMsg>();

	long lastCheckTime = System.currentTimeMillis();
	
	public String getProcessName() {
		return processName;
	}

	/**
	 * 重置灰度标识
	 */
	public void onResetGrid(ServantEndPoint service, int newGridValue) {
		try {
			serviceInfos.setServiceGrid(service, newGridValue);
		} catch (Exception e) {
			TafLoggerCenter.error("onResetGrid error " + objectName + " " + e);
		}
	}	
}
